//
//	Yampp player link tool
//
//        Copyright (c) 2003 Jim Peters <http://uazu.net/>.  Released
//        under the GNU GPL version 2 as published by the Free
//        Software Foundation.  See the file COPYING for details, or
//        visit <http://www.gnu.org/copyleft/gpl.html>.
//
//	For more details on the Yampp MP3 players, see:
//
//	  http://yampp.com/
//
//	On Linux, you need to get the correct FTDI USB drivers for
//	your Linux kernel enabled and compiled in if they aren't
//	already.  These come with modern kernel sources (later 2.4.*),
//	so it is just a configuration option change: enable
//	Experimental Features (under "Code maturity level options"),
//	go to USB, enable USB Serial Convertor Support, enable FTDI,
//	and rebuild, install, reboot.  You might also need to run
//	"/dev/MAKEDEV usb" to create the correct /dev/ttyUSB* device
//	files.  After all this is done, plugging in the yampp USB
//	connection should result in a message appearing in the syslog,
//	something like "FTDI 8U232AM converter now attached to
//	ttyUSB0".  If this quick-howto of mine hasn't helped, you'll
//	need to consult all the relevant Linux HOWTOs for each stage
//	to figure out what you need to do.
//

#define VERSION "0.9.1"

//
//	Editing playlists
//	-----------------
//
//	By default all newly 'put' MP3s are added to playlist 0.  If
//	you want to adjust this, do a 'plget', and edit the resulting
//	playlist.txt file.  You can create new playlists by adding
//	another @name line and cutting/copying/pasting some songs
//	below it.  The song names are for your information only --
//	actually it is only the song numbers that are taken account
//	of.  If there are songs on the yampp that aren't in the
//	playlist, you can add them just by using the song number on a
//	line by itself.  Once you have the playlists as you want them,
//	use 'plput' to load your new playlists into the yampp.
//

//
//	Filenames
//	---------
//
//	When looking for artist and title in a filename, first ylink
//	looks for '--' or '__' or ' ' as a separator.  Otherwise it
//	looks for '-' or '_' or ' ', and takes the remainder to be the
//	title.  When loading an MP3 back off the player, it uses '--'
//	as the separator.
//
	
//
//	Possible improvments
//	--------------------
//
//	- No attempt is made currently to extract information from the
//	MP3 file.  The song-length is set to 5 minutes, and the
//	bit-rate to 120kbps.  This doesn't matter unless you have an
//	LCD on your player.  This info could all be picked up when
//	copying the MP3 data across.
//
//	- At the moment the yampp link is brought up and down each
//	time the 'ylink' command is issued.  It could be left up
//	(i.e. without giving a USB_EXIT_LINK command) between
//	invocations of ylink, with a special 'ylink exit' command to
//	return the yampp to normal operation.
//
//	- 'ylink' loads up the songbase, FAT and playlists every time
//	it runs.  This could be very slow on hard drive yampps.  This
//	information could be cached and stored locally in a temporary
//	file between invocations of 'ylink', reloading only if the
//	yampp has a different root sector to what is expected.
//
//	- It might be useful to do retries if a yampp command fails.
//	However, I've never had this problem, even with 55MB MP3
//	downloads to the player.  Once it is on and listening okay,
//	everything seems to go fine (i.e. sometimes it takes a couple
//	of pings before the yampp is in the right mode and ready, but
//	after that, no problem).
//
//	- It would be good to not rewrite the playlist/etc so many
//	times when putting a list of MP3s onto the device.  But this
//	was judged to be better than losing the lot if there was a
//	communication failure.
//
//	- Maybe add commands to support scripting better, or sort the
//	songbase (or listing) into artist/title order.
//

#include <sys/types.h>                                                    
#include <sys/stat.h>                                                     
#include <fcntl.h>                                                        
#include <termios.h>                                                      
#include <time.h>                                                        
#include <stdio.h>                                                        
#include <stdlib.h>                                                        
#include <stdarg.h>                                                        
#include <string.h>                                                        
#include <errno.h>                                                        
#include <unistd.h>


//
//	Error handling
//

static void 
error(char *fmt, ...) {
   va_list ap;
   va_start(ap, fmt);
   vfprintf(stderr, fmt, ap);
   fprintf(stderr, "\n");
   exit(1);
}

static void 
errorE(char *fmt, ...) {
   int eno= errno;
   va_list ap;
   va_start(ap, fmt);
   vfprintf(stderr, fmt, ap);
   fprintf(stderr, "\n  %s\n", strerror(eno));
   exit(1);
}

static void 
unexpected(char *nam) {
   errorE("Unexpected error from \"%s\":", nam);
}

static void 
warn(char *fmt, ...) {
   va_list ap;
   va_start(ap, fmt);
   vfprintf(stdout, fmt, ap);
   printf("\n");
}

static FILE *pr_out;

static void 
pr(char *fmt, ...) {
   va_list ap;
   va_start(ap, fmt);
   vfprintf(pr_out, fmt, ap);
   fprintf(pr_out, "\n");
}

//
//	Memory handling
//

void *
Alloc(int size) {
   void *vp= calloc(1, size);
   if (!vp) error("Out of memory");
   return vp;
}

#define ALLOC(type) ((type*)Alloc(sizeof(type)))
#define ALLOC_ARR(cnt, type) ((type*)Alloc((cnt) * sizeof(type)))


//
//	Generate a hexdump to the given stream
//

static void 
hexdump(FILE *out, void *dat, int len) {
   unsigned char *p= dat;
   int a;
   int addr;

   for (addr= 0; addr < len; addr += 16) {
      fprintf(out, "%04x  ", addr);
      for (a= 0; a<16; a++) {
	 if (a == 8)
	    fprintf(out, " ");
	 if (a+addr < len) 
	    fprintf(out, " %02x", p[a]);
	 else 
	    fprintf(out, "   ");
      }
      fprintf(out, "  ");

      for (a= 0; a<16; a++) {
	 int ch= (a+addr>=len) ? ' ' : *p++;
	 if (ch < ' ' || ch > '~') ch= '.';

	 if (a == 8)
	    fprintf(out, " ");
	 fprintf(out, "%c", ch);
      }
      fprintf(out, "\n");
   }
}

//
//      Load a file as a single NUL-terminated string into memory.
//      Also saves the loaded length into *lenp if 'lenp' is non-zero,
//      which is useful for binary data.
//

static char *
load_file(char *fnam, int *lenp) {
   int len;
   char *rv;
   FILE *in;

   if (!(in= fopen(fnam, "rb")))        // "b" binary for Windows
      error("Can't open file: %s", fnam);

   if (0 != fseek(in, 0, SEEK_END) ||
       0 > (len= ftell(in)))
      error("Can't find file length: %s", fnam);

   fseek(in, 0, SEEK_SET);

   rv= Alloc(len+1);
   if (1 != fread(rv, len, 1, in) ||
       0 != fclose(in))
      error("Read error on file: %s", fnam);

   rv[len]= 0;
   if (lenp) *lenp= len;
   return rv;
}


//
//	Globals
//

int verb= 0;		// Verbose reporting?
int progress= 1;	// Write progress meter for long writes
int yw_targ;		// Write target sector total, or 0
int yw_curr;		// Current write sector total
int yr_targ;		// Read target sector count, or 0
int yr_curr;		// Current read sector total

//
//	Serial port handling
//

static int port_fd;
static struct termios port_tio_old, port_tio_new;

//
//	Open the USB serial port.  Any failure results in a call to
//	error().  Call this only once if you want the port restored to
//	its previous settings correctly.  An atexit() handler is set
//	up to close the port on exit.  Baud rate is set to 38400.
//	This shouldn't have any effect on an FTDI USB device, I hope.
//

static void 
port_open(char *dev) {
   int fd;
   static void port_atexit();

   port_fd= fd= open(dev, O_RDWR | O_NOCTTY);
   
   if (fd < 0) 
      errorE("Failed to open USB serial port: %s", dev);

   if (0 != tcgetattr(fd, &port_tio_old)) 
      unexpected("tcgetattr");

   memset(&port_tio_new, 0, sizeof(port_tio_new));
   port_tio_new.c_cflag= B38400 | CRTSCTS | CS8 | CLOCAL | CREAD;
   port_tio_new.c_iflag= IGNPAR;
   port_tio_new.c_oflag= 0;
   port_tio_new.c_lflag= 0;      // set input mode (non-canonical, no echo,...)
   port_tio_new.c_cc[VTIME]= 5;  // 0.5 sec time-out
   port_tio_new.c_cc[VMIN]= 0;   // return single bytes if that's all that arrives
   
   if (0 != tcflush(fd, TCIFLUSH))
      unexpected("tcflush");

   if (0 != tcsetattr(fd, TCSANOW, &port_tio_new)) 
      unexpected("tcsetattr");

   atexit(port_atexit);
}

static void 
port_atexit() {
   tcsetattr(port_fd, TCSANOW, &port_tio_old);
   close(port_fd);
}

//
//	Change the port timeout.  (Timeout value is in tenths of a
//	second).
//

void 
port_tmo(int tmo) {
   port_tio_new.c_cc[VTIME]= tmo;

   if (0 != tcsetattr(port_fd, TCSANOW, &port_tio_new)) 
      unexpected("tcsetattr");
}

//
//	Send a chunk of data to the port
//   

static void 
port_write(void *dat, int len) {
   while (len > 0) {
      int rv= write(port_fd, dat, len);
      if (rv < 0 && errno != EAGAIN)
	 errorE("Write error on USB serial port");
      if (rv < 0) continue;
      len -= rv; dat += rv;
   }
}

//
//	Read a chunk of data from the port.  This will time-out if it
//	takes too long, and display a message to stderr if 'warn' is
//	non-zero.  Returns the number of bytes read, which will be
//	less than 'len' if there was a time-out.
//

static int 
port_read_tmo(void *dat, int len, int warn) {
   int len0= len;
   void *dat0= dat;
   while (len > 0) {
      int rv= read(port_fd, dat, len);
      if (rv < 0 && errno != EINTR && errno != EAGAIN)
	 errorE("Read error on USB serial port");
      if (rv < 0) continue;
      if (rv == 0) {
	 int cnt= len0-len;
	 if (warn) {
	    fprintf(stderr, "Read timeout: %d/%d bytes read\n", cnt, len0);
	    hexdump(stderr, dat0, cnt);
	    //hexdump(stderr, dat0, cnt > 16 ? 16 : 0);
	    //if (cnt > 16) fprintf(stderr, " :: \n");
	 }
	 return cnt;
      }
      len -= rv; dat += rv;
   }
   return len0;
}


//
//	Read a chunk of data from the port.  If the read times out,
//	error() is called.  Only returns on success.
//

static void 
port_read(void *dat, int len) {
   if (len != port_read_tmo(dat, len, 1))
      error("ABORT");
}

//
//	yboot (boot-loader) handling
//

//
//	Check status, returns string "yboot1", "yboot2", etc
//

static char *
yboot_status() {
   static char buf[7];

   port_write("?", 1);
   port_read(buf, 6);
   buf[6]= 0;

   if (0 != memcmp(buf, "yboot", 5)) {
      warn("Bad yboot status string:");
      hexdump(stderr, buf, 6);
      error("Abort");
   }

   return buf;
}

//
//	Exit bootloader mode
//

static void 
yboot_esc() {
   port_write("ESC", 3);
}

//
//	Download data
//

static void 
yboot_download(char *dat, int len) {
   char buf[4];
   int bcnt= len >> 7;
   int a;

   if (len % 128 != 0)
      error("yboot_download attempted with length not a multiple of 128");

   buf[0]= '!';
   buf[1]= 'A';
   buf[2]= 'B';
   buf[3]= bcnt;

   port_write(buf, 4);

   for (a= 0; a<bcnt; a++) {
      if (!port_read_tmo(buf, 1, 1) || buf[0] != 'A')
	 error("Expected ACK didn't arrive during yboot download;\n"
	       "%d/%d blocks written", a, bcnt);
      port_write(dat, 128);
      dat += 128;
   }
   if (!port_read_tmo(buf, 1, 1) || buf[0] != 'A') 
      error("Expected final ACK didn't arrive after yboot download");
}

//
//	Yampp player command handling
//      

typedef unsigned int u32;
typedef unsigned short u16;
typedef unsigned char u08;
typedef signed long long int s64;

#include "usb.h"


//
//	Send the yampp player a USB command block.  Format gives list
//	of value types to include in UCB:
//
//	  I 4-byte integer, pulled from arg list
//	  C 1-byte integer, pulled from arg list
//	  _ 4-byte skip
//	  . 1-byte skip
//
//	Possible calls:
//
//	  yampp_ucb("II", USB_TEST_POLL, 'a');
//	  yampp_ucb("II", USB_TEST_POLL, 'b');
//	  yampp_ucb("I", USB_ENTER_BOOT);
//	  yampp_ucb("III", USB_EEPROM_READ, addr, len);
//	  yampp_ucb("III", USB_EEPROM_WRITE, addr, len);
//	  yampp_ucb("III", USB_SECTOR_READ, sector, n_sectors);
//	  yampp_ucb("III", USB_SECTOR_WRITE, sector, n_sectors);
//	  yampp_ucb("IIC", USB_SECTOR_FILL, sector, fill);
//	  yampp_ucb("I", USB_IDENTIFY);			// Gets CF disk info
//	  yampp_ucb("II_I", USB_PLAY, len, total_bits);	// Not sure here
//	  yampp_ucb("II", USB_BUFFER, len);
//	  yampp_ucb("I", USB_STOP);
//	  yampp_ucb("I", USB_ENTER_LINK);	// Call before sending other commands
//	  yampp_ucb("I", USB_EXIT_LINK);	// Call after sending commands
//	  yampp_ucb("ICC", USB_SETVOLUME, left, right);
//	  yampp_ucb("I", USB_BUFFERINFO);
//

static void 
yampp_ucb(char *fmt, ...) {
   char ch;
   u32 ucb[4];
   u08 *cp= (void*)ucb;
   va_list ap;

   ucb[0]= ucb[1]= ucb[2]= ucb[3]= 0;
   va_start(ap, fmt);
   
   while ((ch= *fmt++)) switch (ch) {
    case 'I': *((u32*)cp)= va_arg(ap, int); cp += 4; break;
    case 'C': *cp++= va_arg(ap, int); break;
    case '_': cp += 4; break;
    case '.': cp++; break;
   }

   port_write(&ucb[0], 16);
}

//
//	Sector size is apparently hard-coded to 512 bytes in all
//	firmware.  I originally thought that USB_BUFFERINFO returned
//	the sector size, but instead it appears to be connected with
//	USB playback.  So, sector size is fixed here to 512 bytes.
//

static int s_size= 512;


//
//	Get a response code.  This should normally be ACK, NAK or UNK.
//	If there is a timeout, error() is called.
//

static int 
yampp_read_rc() {
   u08 buf[2];

   if (2 != port_read_tmo(buf, 2, 1))
      error("Timed out waiting for response code");

   return buf[0] + buf[1] * 256;
}


//
//	Enter/exit link mode
//

void 
yampp_link_up() {
   if (verb) warn("Yampp link up ...");

   yampp_ucb("I", USB_ENTER_LINK);	// Call before sending other commands
   if (ACK != yampp_read_rc())
      error("Yampp failed to acknowledge ENTER LINK");
}

void 
yampp_link_down() {
   if (verb) warn("Yampp link down ...");

   yampp_ucb("I", USB_EXIT_LINK);	// Call after sending commands
   if (ACK != yampp_read_rc())
      error("Yampp failed to acknowledge EXIT LINK");
}

//
//	Read data from the CF card.  Returns 0: error, 1: success.
//	This is fast if successful, but if there is an illegal sector,
//	or other error reported from the yampp, it is necessary to
//	wait for a timeout.
//

static int 
yampp_read(char *dat, int num, int cnt) {
   int len, rv;
   int prog= progress && cnt > 64;

   //if (!s_size) yampp_s_size();

   if (verb) {
      warn(cnt == 1 ? 
	   "Yampp read sector %d ..." :
	   "Yampp read sectors %d to %d ...", num, num+cnt-1);
   }

   // Read them 32 sectors at a time
   while (cnt > 0) {
      int cc= cnt > 32 ? 32 : cnt;

      // Progress meter
      if (prog) {
	 if (yr_targ) {
	    fprintf(stderr, "  Reading %d/%d ...    \r", yr_curr, yr_targ);
	 } else {
	    fprintf(stderr, cc == 1 ? 
		    "  Writing sector %d ...    \r" :
		    "  Writing sectors %d to %d ...    \r", 
		    num, num+cc-1);
	 }
	 fflush(stderr);
      }

      // Read it in
      yampp_ucb("III", USB_SECTOR_READ, num, cc);
      len= s_size * cc;
      if (len != (rv= port_read_tmo(dat, len, 0))) {
	 if (rv == 2 && dat[0] == 15 && dat[1] == 0) 
	    return 0;
	 warn("Timeout whilst reading sector data, %d/%d bytes read:", rv, len);
	 hexdump(stderr, dat, rv > 512 ? 512 : rv);
	 if (rv > 512) warn(" ::");
	 error("ABORT");
      }

      if (ACK != yampp_read_rc())
	 error("Sector read has missing ACK");

      num += cc;
      cnt -= cc;
      dat += len;
      yr_curr += cc;
   }

   //	// Read all in one go
   //   yampp_ucb("III", USB_SECTOR_READ, num, cnt);
   //   len= cnt * s_size;
   //   if (len != (rv= port_read_tmo(dat, len, 0))) {
   //      if (rv == 2 && dat[0] == 15 && dat[1] == 0) 
   //	 return 0;
   //      warn("Timeout whilst reading sector data, %d/%d bytes read:", rv, len);
   //      hexdump(stderr, dat, rv > 512 ? 512 : rv);
   //      if (rv > 512) warn(" ::");
   //      error("ABORT");
   //   }
   //   
   //   if (ACK != yampp_read_rc())
   //      error("Sector read has missing ACK");

   // Clean up progress meter
   if (prog) {
      fprintf(stderr, "                                        \r");
      fflush(stderr);
   }
   
   return 1;
}
   
//
//	Write data to the CF card.  Calls error() if there is a
//	problem.
//

static void 
yampp_write(char *dat, int num, int cnt) {
   int len, rv;
   int cnt0= cnt;
   int prog= progress && cnt > 64;

   //if (!s_size) yampp_s_size();

   // Verbose reporting
   if (verb) {
      warn(cnt == 1 ? 
	   "Yampp write sector %d ..." :
	   "Yampp write sectors %d to %d ...", 
	   num, num+cnt-1);
   }
   
   // Loop writing 32-sector chunks at a time (16KB) instead of
   // writing it all in one go.
   while (cnt > 0) {
      int cc= cnt > 32 ? 32 : cnt;

      // Progress messages
      if (prog) {
	 if (yw_targ) {
	    fprintf(stderr, "  Writing %d/%d ...    \r",  yw_curr, yw_targ);
	 } else {
	    fprintf(stderr, cc == 1 ? 
		    "  Writing sector %d ...    \r" :
		    "  Writing sectors %d to %d ...    \r", 
		    num, num+cc-1);
	 }
	 fflush(stderr);
      }

      // Write it
      yampp_ucb("III", USB_SECTOR_WRITE, num, cc);
      len= cc * s_size;
      port_write(dat, len);

      if (ACK != (rv= yampp_read_rc())) 
	 error("USB_SECTOR_WRITE failed with response code %d", rv);

      cnt -= cc;
      num += cc;
      dat += len;
      
      yw_curr += cc;
   }

   // Clean up progress meter
   if (prog) {
      fprintf(stderr, "                                        \r");
      fflush(stderr);
   }
}
   
//
//	*** From here on down, assume that the sector size will be 512
//	bytes ***
//

//
//	YADL format handling 
//

#include "yadl.h"
#define YADL_RSH_SEC_CLU 9
#define YADL_RSH_BYT_SEC 9

YADL_ROOT_SECTOR root;
int root_valid;				// Does the root sector contain a valid YADL record?
YADL_SONG_BASE_ENTRY *songbase= 0;	// Valid data, or 0 pointer
u32 *fat= 0;				// Valid data, or 0 pointer
YADL_PLAY_LIST_ENTRY *playlist;		// Valid data, or 0 pointer
u16 **plarr;				// Array of arrays of Songbase indexes
char *pldat;				// Combined data for playlist
int pldat_sect;				// Number of sectors required for playlist data

//
//	Load the root sector to the 'root' global structure
//

void 
load_root_raw() {
   if (!yampp_read((void*)&root, 0, 1)) 
      error("Can't load yampp root sector");

   root_valid= (0 == memcmp("YADL", root.id, 4));
}

void 
load_root() {
   load_root_raw();

   if (!root_valid)
      error("Root sector is invalid.  Please allow the yampp firmware to initialise\n"
	    " the root sector, and then do 'ylink format'.  To force recreation\n"
	    " of the root sector, use 'ylink zeroroot'.");
}

//
//	Save the root sector from the 'root' global structure.
//	'final' should be 1 when writing the final root sector that
//	indicates that everything is straight.  0 writes the 'mounted'
//	flag, indicating unclean.
//

void 
save_root(int final) {
   struct tm *tm;
   time_t now= time(0);

   root.unmount_state= final ? YADL_STATE_OK : YADL_STATE_MOUNTED;

   // Update timestamp
   tm= localtime(&now);
   root.t_year= 1900 + tm->tm_year;
   root.t_month= 1 + tm->tm_mon;
   root.t_day= tm->tm_mday;
   root.t_hour= tm->tm_hour;
   root.t_minute= tm->tm_min;
   root.t_second= tm->tm_sec;

   // Write it
   warn("Writing root sector ...");
   yampp_write((void*)&root, 0, 1);
}

//
//	Load/save the songbase
//

void 
load_songbase() {
   int cnt= root.fat_start - root.songbase_start;

   if (songbase) free(songbase);
   songbase= Alloc(cnt * SECTORSIZE);

   if (!yampp_read((char*)songbase, root.songbase_start, cnt))
      error("Failed to read songbase");
}

void 
save_songbase() {
   int cnt= root.fat_start - root.songbase_start;

   if (!songbase)
      error("Attempt to save songbase when not created");

   warn("Writing songbase ...");
   yampp_write((char*)songbase, root.songbase_start, cnt);
}   
   

//
//	Load/save the FAT
//   

void 
load_fat() {
   int cnt= root.sector_cluster_offset - root.fat_start;

   if (fat) free(fat);
   fat= Alloc(cnt * SECTORSIZE);

   if (!yampp_read((char*)fat, root.fat_start, cnt))
      error("Failed to read FAT");
}
   
void 
save_fat() {
   int cnt= root.sector_cluster_offset - root.fat_start;

   if (!fat)
      error("Attempt to save FAT when not created");

   warn("Writing FAT ...");
   yampp_write((char*)fat, root.fat_start, cnt);
}   

//
//	Delete all non-existant songs from the playlists
//

void 
cleanup_playlist(int info) {
   int a, b, c;
   int n_pl= root.playlist_qty;

   for (a= 0; a<n_pl; a++) {
      int n_ent= playlist[a].entry_qty;
      for (b= c= 0; b<n_ent; b++) {
	 int ii= plarr[a][b];
	 if (ii >= root.songbase_qty ||
	     songbase[ii].state != 16) {
	    if (info) warn("Warning: deleting non-existant song %d from playlist %d", ii, a);
	 } else {
	    plarr[a][c++]= ii;
	 }
      }
      playlist[a].entry_qty= c;
   }
}

//
//	Load playlist and fix up any problems with it
//

void 
load_playlist() {
   char *dat;
   YADL_PLAY_LIST_ENTRY *pl;
   int cnt;
   int n_pl;
   u32 curr;
   int a, b;
   
   if (playlist) { free(playlist); playlist= 0; }
   
   if (root.playlist_start != YADL_FAT_INVALID && 
       root.playlist_start != 0)
      error("Can't handle playlists if they're not in the first cluster");
   
   // Check for missing playlist and create a blank one instead
   if (root.playlist_qty < 1 ||
       root.playlist_start >= root.fat_qty ||
       root.playlist_file_length < 64) {
      warn("Creating default playlist");
      root.playlist_qty= 1;
      root.playlist_start= 0;
      root.playlist_file_length= 1024;
      root.playlist_first= 0;
      if (fat[0] != YADL_FAT_EMPTY)
	 error("Trying to create playlist, but can't because first cluster is occupied");
      fat[0]= YADL_FAT_LAST;
      playlist= ALLOC(YADL_PLAY_LIST_ENTRY);
      strcpy(playlist->name, "default");
      playlist[0].next_playlist= -1;
      plarr= ALLOC(u16*);
      plarr[0]= ALLOC_ARR(1, u16);
   } else {
      // Load up the playlist that is already there and 'normalize' it
      cnt= (root.playlist_file_length + 511) / 512;
      dat= Alloc(cnt * 512);
      
      if (!yampp_read(dat, root.sector_cluster_offset + 512 * root.playlist_start, cnt))
	 error("Failed to read Playlist");

      // Straighten it all out
      pl= (void*)dat;
      n_pl= 0;
      curr= root.playlist_first;
      while (curr != 65535) {
	 n_pl++;
	 curr= pl[curr].next_playlist;
	 if (curr == root.playlist_first) break;
      }

      // Hack to fix up playlists from yamppLink, i.e. playlists with
      // all 'next_playlist' link set to -1, which appear to just work
      // on the root.playlist_qty; I don't want to ignore all the
      // next_playlist values now I've got that coded up, but it
      // appears that they are not used by yamppLink or the firmware
      // at present.
      if (root.playlist_qty > n_pl) {
	 warn("Fixing up yamppLink playlists ...");
	 root.playlist_first= 0;
	 for (a= 0; a<root.playlist_qty; a++) 
	    pl[a].next_playlist= (a+1 == root.playlist_qty) ? 65535 : a+1;
	 n_pl= root.playlist_qty;
      }

      // Allocate space to store them and copy them over in order
      playlist= ALLOC_ARR(n_pl, YADL_PLAY_LIST_ENTRY);
      plarr= ALLOC_ARR(n_pl, u16*);
      
      curr= root.playlist_first;
      for (a= 0; a<n_pl; a++) {
	 memcpy(playlist + a, pl + curr, sizeof(*pl));
	 curr= pl[curr].next_playlist;
	 playlist[a].next_playlist= (a == n_pl-1) ? -1 : a+1;
	 plarr[a]= ALLOC_ARR(playlist[a].entry_qty, u16);
	 memcpy(plarr[a], dat + playlist[a].offset, 
		playlist[a].entry_qty * sizeof(u16));
	 if (curr == root.playlist_first) break;
      }
      root.playlist_qty= n_pl;

      // Delete loaded data
      free(dat); dat= 0;
      
      // Check for non-existent songs in playlists and delete them
      cleanup_playlist(1);
   }		  
}			     

//
//	Fix up any FAT errors (or just warn if 'fix' is 0)
//

void 
fix_fat(int fix) {
   int n_fat= root.fat_qty;
   char *map= Alloc(n_fat);
   int a;

   // Mark all clusters in use
   for (a= 0; a<n_fat; a++) 
      map[a]= fat[a] <= YADL_FAT_LAST;

   // Clear clusters that are accounted for in various ways
   for (a= -1; a<root.songbase_qty; a++) {
      u32 curr= (a<0) ? root.playlist_start : songbase[a].start;
      do {
	 if (curr >= n_fat) {
	    warn("WARNING: Broken FAT chain for song %d: accessing cluster %d", a, curr);
	    break;
	 }
	 map[curr]= 0;
	 curr= fat[curr];
      } while (curr < YADL_FAT_LAST);
   }

   for (a= 0; a<n_fat; a++) if (map[a]) {
      if (fix) {
	 fat[a]= YADL_FAT_EMPTY;
      } else {
	 warn("WARNING: Allocated but unused FAT entry: %d", a);
      }
   }

   free(map);
}

//
//	Prepare playlists for saving; creates 'pldat', writes various
//	info into root sector and updates songbase with reference
//	counts/etc
//

void 
prep_playlist() {
   int a, b;
   int n_pl= root.playlist_qty;
   int tlen= 0;
   char *p;

   // Calculate size of pldat
   for (a= 0; a<n_pl; a++) {
      tlen += sizeof(playlist[0]);
      tlen += playlist[a].entry_qty * sizeof(u16);
   }
   root.playlist_file_length= tlen;

   // Get all the songbase reference counts right, and calculate
   // playlist total times
   for (a= 0; a<root.songbase_qty; a++)
      songbase[a].ref_ctr= 0;
   for (a= 0; a<n_pl; a++) {
      int total= 0;		// Total time in seconds
      for (b= 0; b<playlist[a].entry_qty; b++) {
	 int ii= plarr[a][b];
	 if (ii >= 0 && ii < root.songbase_qty) {
	    songbase[ii].ref_ctr++;
	    total += songbase[ii].play_time;
	 }
      }
      playlist[a].length_hours= total / 3600;
      playlist[a].length_minutes= total % 3600 / 60;
      playlist[a].length_seconds= total % 60;
   }

   // Allocate the playlist file
   pldat_sect= (tlen + 511) / 512;
   if (pldat) free(pldat);
   pldat= Alloc(pldat_sect * 512);
   
   // Copy list data over and set playlist offsets
   p= pldat + n_pl * sizeof(playlist[0]);
   for (a= 0; a<n_pl; a++) {
      int len= playlist[a].entry_qty * sizeof(u16);
      playlist[a].offset= p-pldat;
      memcpy(p, plarr[a], len);
      p += len;
   }

   // Finally insert the playlist structures themselves
   memcpy(pldat, playlist, n_pl * sizeof(playlist[0]));
}


//
//	Save playlist (prep_playlist() should have been called first)
//   

void 
save_playlist() {
   warn("Writing playlist ...");
   yampp_write(pldat, root.sector_cluster_offset + 512 * root.playlist_start, pldat_sect);
}

//
//	Load everything, and check everything and fix things up as
//	necessary
//   

void 
load_all() {
   load_root();
   load_songbase();
   load_fat();
   load_playlist();

   fix_fat(0);		// Generate warnings
}

//
//	Save everything
//

void 
save_all() {
   prep_playlist();
   save_root(0);	// Save in progress
   save_fat();
   save_songbase();
   save_playlist();
   save_root(1);	// Save complete
}
			     
//
//	Write playlist out to 'playlist.txt' in the current directory
//

void 
write_playlist() {
   int a, b;

   pr_out= fopen("playlist.txt", "w");
   if (!pr_out) error("Can't create file: 'playlist.txt'");
   
   for (a= 0; a<root.playlist_qty; a++) {
      pr("@%s", playlist[a].name);
      for (b= 0; b<playlist[a].entry_qty; b++) {
	 int ii= plarr[a][b];
	 if (ii < 0 || ii >= root.songbase_qty) continue;
	 pr("%d\t(%s / %s)", ii, songbase[ii].artist, songbase[ii].title);
      }
      pr("");
   }
   
   if (0 != fclose(pr_out)) 	
      errorE("Error writing to 'playlist.txt'");
   pr_out= 0;
}

//
//	Read 'playlist.txt' in and create playlist structures from it
//

void 
read_playlist() {
   char *dat= load_file("playlist.txt", 0);
   char *datend= strchr(dat, 0);
   char *p, *q, ch;
   int n_pl;
   int a, b;

   // Free everything already existing first
   if (playlist) {
      for (a= 0; a<root.playlist_qty; a++) 
	 if (plarr && plarr[a]) free(plarr[a]);
      if (plarr) free(plarr);
      free(playlist);
      plarr= 0;
      playlist= 0;
      root.playlist_qty= 0;
   }

   // Convert all '\n' characters to NULs, strip leading/trailing white
   // space
   for (p= dat; p < datend; p++) 
      if (*p == '\n') 
	 *p= 0;
   for (p= dat; p < datend; p++) {
      while (isspace(*p)) *p++= 0;
      while (*p) p++;
      for (q= p; q > dat && isspace(q[-1]);) *q--= 0;
   }

   // Count the number of playlists in this file
   n_pl= 0;
   for (p= dat; p < datend; p= 1+strchr(p, 0))
      if (*p == '@') n_pl++;

   // Fill in the playlist structures and counts
   root.playlist_qty= n_pl;
   playlist= ALLOC_ARR(n_pl, YADL_PLAY_LIST_ENTRY);
   a= -1;
   for (p= dat; p < datend; p= 1+strchr(p, 0)) {
      if (*p == '@') {
	 p++; a++;
	 if (strlen(p)+1 > sizeof(playlist[a].name))
	    error("Playlist name \"%s\" is too long", p);
	 strcpy(playlist[a].name, p);
	 playlist[a].next_playlist= (a == n_pl-1) ? -1 : a+1;
	 continue;
      }
      if (isdigit(*p)) {
	 if (a < 0) error("Error: 'playlist.txt' contains digits before first '@' line");
	 playlist[a].entry_qty++;
	 continue;
      }
      if (*p) warn("Illegal line in 'playlist.txt', ignored:\n  %s", p);
   }

   // Fill in the lists themselves
   plarr= ALLOC_ARR(n_pl, u16*);
   for (a= 0; a<n_pl; a++)
      plarr[a]= ALLOC_ARR(playlist[a].entry_qty, u16);
   a= -1;
   for (p= dat; p < datend; p= 1+strchr(p, 0)) {
      if (*p == '@') {
	 a++; b= 0;
	 continue;
      } 
      if (isdigit(*p)) {
	 int ii= strtol(p, 0, 10);
	 if (ii < 0 || ii >= root.songbase_qty || 
	     songbase[ii].state != 16) {
	    warn("Song number %d does not exist on the player; ignoring", ii);
	    playlist[a].entry_qty--;
	 } else {
	    plarr[a][b++]= ii;
	 }
      }
   }

   // That should be all done
   free(dat);
}
   

//
//	Dump the contents of the root sector out to the current
//	'pr_out' FILE*
//

void 
dump_root() {
   int a;

   // Double-check root_valid
   root_valid= (0 == memcmp("YADL", root.id, 4));

   if (!root_valid) {
      pr("  Bad root sector ID: %02x %02x %02x %02x", 
	 root.id[0], root.id[1], root.id[2], root.id[3]);
      return;
   }
   
   pr("  Disk size: %d sectors (%dKB)", 
      root.disk_size, root.disk_size/2);

   a= root.unmount_state;
   pr("  Disk state: %d (%s)", a,
      a == YADL_STATE_OK ? "YADL_STATE_OK" :
      a == YADL_STATE_MOUNTED ? "YADL_STATE_MOUNTED" :
      a == YADL_STATE_NEW ? "YADL_STATE_NEW" :
      "??");

   a= root.media_type;
   pr("  Media type: %d (%s)", a,
      a == YADL_MEDIA_DISK ? "YADL_MEDIA_DISK" :
      a == YADL_MEDIA_FLASH ? "YADL_MEDIA_FLASH" :
      a == YADL_MEDIA_UNKNOWN ? "YADL_MEDIA_UNKNOWN" :
      "??");

   pr("  Disk format version: %d.%d", 
      root.disk_version_major,
      root.disk_version_minor);
   
   pr("  Yampp model: %s", 
      root.model);
   
   pr("  Disk info: %s", 
      root.disk_info);

   pr("  Firmware build timestamp: %s",
      root.build_timestamp);

   pr("  Firmware version: %d.%d",
      root.firmware_version_major,
      root.firmware_version_minor);

   pr("  Disk identifier: \"%s\"",
      root.identifier);

   pr("  Last access: %04d-%02d-%02d %02d:%02d:%02d",
      root.t_year, root.t_month, root.t_day,
      root.t_hour, root.t_minute, root.t_second);
   
   pr("  Cluster offset: sector %d, %d free clusters",
      root.sector_cluster_offset,
      root.free_cluster_count);

   if (root.wasteland_length)
      pr("  Wasteland sectors: %d to %d (%dKB)",
	 root.wasteland_start,
	 root.wasteland_start + root.wasteland_length - 1,
	 root.wasteland_length/2);

   pr("  Songbase: from sector %d, %d entries",
      root.songbase_start,
      root.songbase_qty);

   pr("  Indexes: artist %d, title %d",
      root.first_artist,
      root.first_title);

   pr("  FAT area: from sector %d, %d entries",
      root.fat_start,
      root.fat_qty);
 
   pr("  Playlists: from cluster %d, %d bytes, %d playlists, first is %d",
      root.playlist_start,        
      root.playlist_file_length,
      root.playlist_qty,
      root.playlist_first);

   pr("  Text: from sector %d, %d entries, version %d.%d",
      root.text_data,
      root.text_qty,
      root.text_version_major,
      root.text_version_minor);

   pr("  System sounds: from cluster %d, %d entries, version %d.%d",
      root.snd_system,
      root.snd_qty,
      root.snd_version_major,
      root.snd_version_minor);

   pr("  Options: %d %d %d %d %d %d %d %d %d %d",
      root.option_00,
      root.option_01,
      root.option_02,
      root.option_03,
      root.option_04,
      root.option_05,
      root.option_06,
      root.option_07,
      root.option_08,
      root.option_09);
}

//
//	Dump Songbase to 'pr_out'
//

void 
dump_songbase() {
   YADL_SONG_BASE_ENTRY *sb;
   int a;
   int n_sb= root.songbase_qty;
   int inuse= 0;

   if (!songbase) return;

   for (a= 0, sb= songbase; a<n_sb; a++, sb++) 
      if (sb->state == 16) inuse++;

   pr("  %d/%d songbase slots in use", inuse, n_sb);

   for (a= 0, sb= songbase; a<n_sb; a++, sb++) if (sb->state) {
      pr("  Songbase entry: %d, state %d (%s)", 
	 a, sb->state,
	 sb->state == 0 ? "FREE" :
	 sb->state == 16 ? "IN USE" :
	 "??");
      pr("    Artist: %s", sb->artist);
      pr("    Title: %s", sb->title);
      pr("    Cluster: %d, %d bytes long", sb->start, sb->length);
      pr("    Reference count: %d", sb->ref_ctr);
      pr("    Play time: %d seconds", sb->play_time);
      pr("    Initial offset: %d bytes", sb->start_offset);
      pr("    List links: artist %d, grouped %d, title %d",
	 sb->next_artist,
	 sb->next_artist_grouped,
	 sb->next_title);
   }
}

//
//	Dump FAT to 'pr_out'
//
   
void 
dump_fat() {
   int a;

   if (!fat) return;

   for (a= 0; a<root.fat_qty; a++) {
      u32 val= fat[a];

      if ((a&7) == 0) 
	 fprintf(pr_out, "  %06x:", a);

      switch (val) {
       case YADL_FAT_INVALID:
	  fprintf(pr_out, " Invalid"); break;
       case YADL_FAT_EMPTY:
	  fprintf(pr_out, " (free) "); break;
       case YADL_FAT_FAULTY:
	  fprintf(pr_out, " *FAULT*"); break;
       case YADL_FAT_SYSTEM:
	  fprintf(pr_out, " System!"); break;
       case YADL_FAT_LAST:
	  fprintf(pr_out, " -->LAST"); break;
       default:
	  if (val == a+1)
	     fprintf(pr_out, " ------>"); 
	  else 
	     fprintf(pr_out, " %07x", val);
      }
      
      if ((a&7) == 7 || a == root.fat_qty-1) 
	 fprintf(pr_out, "\n");
   }
}

//
//	Dump playlists to 'pr_out'
//

void 
dump_playlist() {
   int a, b;
   for (a= 0; a<root.playlist_qty; a++) {
      char buf[32];
      int col;
      pr("Playlist number %d:", a);
      pr("  Name: %s", playlist[a].name);
      pr("  Length: %d:%02d:%02d", playlist[a].length_hours, 
	 playlist[a].length_minutes, playlist[a].length_seconds);
      pr("  System sound: %d/%d", 
	 playlist[a].system_sound_start, 
	 playlist[a].system_sound_length);
      pr("  Entries: %d", playlist[a].entry_qty);
      fprintf(pr_out, "   "); col= 3;
      for (b= 0; b<playlist[a].entry_qty; b++) {
	 int len= sprintf(buf, " %d", plarr[a][b]);
	 col += len;
	 if (col > 72) {
	    fprintf(pr_out, "\n   %s", buf);
	    col= 3 + len;
	 } else {
	    fprintf(pr_out, "%s", buf);
	 }
      }
      pr("");
   }
}
      

//
//	List all the songs on the player to 'pr_out'
//

void 
list_songs() {
   YADL_SONG_BASE_ENTRY *sb;
   int a, b;

   if (!songbase) return;

   // Give summary of usage if FAT is loaded
   if (fat) {
      int total= 0;
      int empty= 0;
      
      for (a= 0; a<root.fat_qty; a++) {
	 u32 val= fat[a];
	 if (val == YADL_FAT_EMPTY)
	    empty++;
	 total++;
      }
      
      pr("Disk %.1f%% free (%dKB), %d/%d clusters in use",
	 (100.0 * empty) / total, empty*256,
      	 total-empty, total);
   }

   for (a= 0, sb= songbase; a<root.songbase_qty; a++, sb++) {
      if (sb->state != 16) continue;
      pr("%4d: %s, \"%s\" (%d:%02d)",
	 a, sb->artist, sb->title,
	 sb->play_time / 60,
	 sb->play_time % 60);
   }

   if (playlist) {
      char buf[32];
      int col;
      int n_pl= root.playlist_qty;
      for (a= 0; a<n_pl; a++) {
	 col= sprintf(buf, "Playlist %d:", a);
	 fprintf(pr_out, "%s", buf);
	 for (b= 0; b<playlist[a].entry_qty; b++) {
	    int len= sprintf(buf, " %d", plarr[a][b]);
	    col += len;
	    if (col > 72) {
	       fprintf(pr_out, "\n %s", buf);
	       col= 1 + len;
	    } else {
	       fprintf(pr_out, "%s", buf);
	    }
	 }
	 pr("");
      }
   }
}   


//
//	Initialise the media based on the root sector already set up
//	by the firmware.  Most of this routine is based on code
//	provided by Uli (Ulrich Behrenbeck).
//

void 
format_disk(char *name) {
   YADL_ROOT_SECTOR rs;
   int num_sectors;
   int num_clusters;
   int fat_size_bytes;
   int song_base_entries;
   int song_base_sectors;
   int fat_size_sectors;
   s64 rest, used;
   int a;

   if (strlen(name) + 1 > sizeof(root.identifier))
      error("Name provided for this disk is too long: \"%s\"", name);

   load_root_raw();

   if (!root_valid)
      error("Bad root sector.  Please allow the firmware to initialise the disk root\n"
	    "  sector before attempting to format the disk from ylink");

   memset(&rs, 0, sizeof(rs));
   memcpy(&rs.id, "YADL", 4);

   rs.unmount_state          = YADL_STATE_MOUNTED;
   rs.media_type             = root.media_type;
   rs.disk_size              = root.disk_size;
   rs.disk_version_major     = root.disk_version_major;
   rs.disk_version_minor     = root.disk_version_minor;
   rs.firmware_version_major = root.firmware_version_major;
   rs.firmware_version_minor = root.firmware_version_minor;
   
   strcpy((char*) rs.disk_info,        (char*) root.disk_info);
   strcpy((char*) rs.model,            (char*) root.model);
   strcpy((char*) rs.build_timestamp,  (char*) root.build_timestamp);
   strcpy((char*) rs.identifier,       (char*) name);
  
   // ready with copying, start calculations
   num_sectors         = rs.disk_size;
   num_clusters        = rs.disk_size >> YADL_RSH_SEC_CLU;
   fat_size_bytes      = num_clusters * sizeof(u32);
   song_base_entries   = ((num_clusters >> 3) & 0xFFFFFFF0) + 10;

   // no more than 0xffff songs allowed
   if (song_base_entries >= 0x10000)
      song_base_entries = 0xffff;

   song_base_sectors = song_base_entries * sizeof(YADL_SONG_BASE_ENTRY) / SECTORSIZE;
   if ((song_base_entries % (SECTORSIZE / sizeof(YADL_SONG_BASE_ENTRY))) != 0)
      song_base_sectors++;

   fat_size_sectors = fat_size_bytes >> YADL_RSH_BYT_SEC;
   if ((fat_size_bytes % SECTORSIZE) != 0)
      fat_size_sectors++;

   // now do sanity check & control the spare room at the end of disk
   rest = num_sectors;
   used = num_clusters * YADL_SECTORS_PER_CLUSTER;

   used += 5;                  // for root, 2 index sectors, appl storage and scratch sector
   used += fat_size_sectors;   // add the FAT here
   used += song_base_sectors;  // for the song base

   rest = rest - used;
   
   while (rest < 0) {
      rest += YADL_SECTORS_PER_CLUSTER;
      num_clusters--;
   }
   
   if (rest != 0) {
      rs.wasteland_length = (u16) rest;
      rs.wasteland_start  = rs.disk_size - (u32) rest;
   }
   rs.songbase_start = YADL_SEC_SONG_BASE;
   rs.songbase_qty   = (u16) song_base_entries;

   rs.fat_start = YADL_SEC_SONG_BASE + (u16) song_base_sectors;
   rs.fat_qty   = num_clusters;
   rs.sector_cluster_offset = 
      rs.fat_start + fat_size_sectors;  // FAT begin+ FAT size = begin of data

   rs.free_cluster_count = num_clusters;

   memcpy(&root, &rs, sizeof(root));

   // Write root sector
   save_root(1);

   // Init sectors 1-4 and songbase
   {
      char tmp[SECTORSIZE];

      // Preset title and artist indices to 0xFFFF
      memset(tmp, 0xFF, sizeof(tmp));
      yampp_write(tmp, YADL_SEC_ARTIST_INDEX, 1);
      yampp_write(tmp, YADL_SEC_TITLE_INDEX, 1);

      // Reset appl_storage and scratch
      memset(tmp, 0, sizeof(tmp));
      yampp_write(tmp, YADL_SEC_APPL_STORAGE, 1);
      yampp_write(tmp, YADL_SEC_SCRATCH, 1);

      // Now clear the songbase
      for (a= 0; a < song_base_sectors; a++) 
	 yampp_write(tmp, YADL_SEC_SONG_BASE+a, 1);
   }

   // Init FAT
   {
      int fat_size_bytes= fat_size_sectors * SECTORSIZE;
      u32 *fat= Alloc(fat_size_bytes);

      memset(fat, 0xFF, fat_size_bytes);

      for (a= 0; a < num_clusters; a++)
	 fat[a]= YADL_FAT_EMPTY;

      yampp_write((char*)fat, root.fat_start, fat_size_sectors);
      free(fat);
   }
}

//
//	Integer argument conversion
//

int 
intarg(char *str) {
   char *end;
   int rv= strtol(str, &end, 10);
   if (end == str) error("Expecting an integer argument: %s", str);
   while (isspace(*end)) end++;
   if (*end) error("Rubbish after the integer argument: %s", str);
   return rv;
}

//
//	Get a song from the player and save it
//

void 
get_song(int ii) {
   char fnam[256];
   FILE *out, *tmp;
   char *buf;
   int len;
   u32 curr;

   if (ii < 0 || ii >= root.songbase_qty ||
       songbase[ii].state != 16)
      error("Song %d doesn't exist", ii);
   
   sprintf(fnam, "%s--%s.mp3", songbase[ii].artist, songbase[ii].title);
   tmp= fopen(fnam, "r");
   if (tmp) error("File '%s' already exists -- ABORTING", fnam);
   
   out= fopen(fnam, "wb");		// "b" just in case someone ports to Windows
   if (!out) errorE("Can't create file '%s'", fnam);
   
   buf= Alloc(512*512);
   len= songbase[ii].length;
   curr= songbase[ii].start;
   
   yr_targ= (len+511)/512;
   yr_curr= 0;

   while (len > 0) {
      // Load up as many sectors of the cluster as we actually need
      int cnt= (len + 511) >> 9;
      if (curr >= YADL_FAT_LAST)
	 error("FAT error -- the FAT chain for this song is incomplete");
      if (!yampp_read(buf, root.sector_cluster_offset + 512 * curr, 
		      cnt > 512 ? 512 : cnt)) {
	 fclose(out);
	 unlink(fnam);
	 error("Problem reading data for cluster %d", curr);
      }
      cnt= (len > 512*512) ? 512*512 : len;
      if (1 != fwrite(buf, cnt, 1, out))
	 errorE("Write error on file '%s'", fnam);
      len -= cnt;
      curr= fat[curr];
   }
   
   if (0 != fclose(out))
      errorE("Error writing file '%s'", fnam);

   free(buf);
}

//
//	Put a song file from the computer to the yampp player
//

void 
put_song(char *fnam) {
   char *p, *q;
   char *fn0= fnam;
   char *fn1, *fn2, *fn3;
   int sn;
   YADL_SONG_BASE_ENTRY *sb;
   int len;		// Filelength
   int len_clust;	// Filelength in clusters
   int a, b;
   char *buf;
   u32 curr;		// Current cluster number
   FILE *in;

   // Look for free slot
   for (sn= 0; sn<root.songbase_qty; sn++)
      if (songbase[sn].state == 0) break;
   if (sn >= root.songbase_qty)
      error("Player Songbase data structure is full");
   sb= songbase + sn;

   // Open the file and check the length
   in= fopen(fnam, "rb");		// "b" in case this is ever ported to Windows
   if (!in) errorE("Can't open file '%s'", fnam);
   if (0 != fseek(in, 0, SEEK_END)) unexpected("fseek");
   if (-1 == (len= ftell(in))) unexpected("ftell");
   if (0 != fseek(in, 0, SEEK_SET)) unexpected("fseek");

   // Check there is space in the FAT
   len_clust= (len + 512*512-1) / (512*512);
   for (a= b= 0; b < len_clust && a<root.fat_qty; a++)
      if (fat[a] == YADL_FAT_EMPTY) 
	 b++;
   if (b < len_clust)
      error("Not enough free space for this song");

   // Look for the artist and title in the filename
   for (p= fnam; *p; p++) if (*p == '/') fn0= p+1;
   fn3= strchr(fn0, 0);
   if (fn3-fn0 >= 4 && 0 == strcasecmp(".mp3", fn3-4))
      fn3 -= 4;
   
   // Scan for a double-character break-point between artist and title
   fn1= fn2= 0;
   for (p= fn0; p<fn3-1; p++) 
      if (p[0] == p[1] &&
	  (p[0] == '-' || p[0] == '_' || p[0] == ' ')) {
	 fn1= p; 
	 fn2= p+2;
	 break;
      }
   
   // Scan for single-char break-points otherwise
   if (!fn1 && (p= strchr(fn0, '-')) && p < fn3-1) {
      fn1= p; fn2= p+1;
   }
   if (!fn1 && (p= strchr(fn0, '_')) && p < fn3-1) {
      fn1= p; fn2= p+1;
   }
   if (!fn1 && (p= strchr(fn0, ' ')) && p < fn3-1) {
      fn1= p; fn2= p+1;
   }
   
   // Last resort, put artist as 'unknown'
   if (!fn1) {
      fn2= fn0;
      fn0= "UNKNOWN";
      fn1= strchr(fn0, 0);
   }

   // Fill in the songbase
   memset(sb, 0, sizeof(*sb));
   if (fn1-fn0+1 > sizeof(sb->artist))
      fn1= fn0 + sizeof(sb->artist) - 1;
   memcpy(sb->artist, fn0, fn1-fn0);
   if (fn3-fn2+1 > sizeof(sb->title))
      fn3= fn2 + sizeof(sb->title) - 1;
   memcpy(sb->title, fn2, fn3-fn2);
   sb->play_time= 300;		// Don't know how to calculate this, so put in 5 minutes
   sb->bitrate= 120;		// Don't know how to calculate this, so put in 120kbps
   sb->state= 16;
   sb->length= len;
   sb->start= ~0;

   // Write data to device and fill in FAT
   yw_targ= (len + 511) / 512;
   yw_curr= 0;
   buf= Alloc(512*512);
   curr= ~0;
   while (len > 0) {
      u32 prev= curr;
      int cnt, cnt_sect;

      curr= 0;
      while (curr < root.fat_qty
	     && fat[curr] != YADL_FAT_EMPTY) curr++;
      if (sb->start == ~0) sb->start= curr;
      fat[curr]= YADL_FAT_LAST;
      if (prev != ~0) fat[prev]= curr;
      
      cnt= (len > 512*512) ? 512*512 : len;
      cnt_sect= (cnt+511) / 512;
      
      if (1 != fread(buf, cnt, 1, in))
	 errorE("Read error on file '%s'", fnam);
      yampp_write(buf, root.sector_cluster_offset + 512 * curr, cnt_sect);
      
      len -= cnt;
   }
   free(buf);
   fclose(in);
   yw_targ= 0;

   // Add the song to playlist zero
   {
      int cnt= playlist[0].entry_qty;
      u16 *tmp= ALLOC_ARR(cnt+1, u16);
      memcpy(tmp, plarr[0], cnt * sizeof(u16));
      tmp[cnt++]= sn;
      free(plarr[0]);
      plarr[0]= tmp;
      playlist[0].entry_qty= cnt;
   }

   // Everything done now
   pr("Song written as song %d:", sn);
   pr("  Artist: %s", sb->artist);
   pr("  Title: %s", sb->title);
}

//
//	Delete a song from the yampp, and any associated playlist
//	entries
//   
   
void 
del_song(int sn) {
   YADL_SONG_BASE_ENTRY *sb;
   int a, b;

   if (sn < 0 || sn >= root.songbase_qty ||
       (sb= &songbase[sn])->state != 16)
      error("Song %d does not exist", sn);

   // Release FAT entries
   {
      u32 curr= sb->start;
      u32 prev;
      do {
	 if (curr >= root.fat_qty) {
	    warn("Bad cluster number %d ignored", curr);
	    break;
	 }
	 prev= curr;
	 curr= fat[prev];
	 fat[prev]= YADL_FAT_EMPTY;
      } while (curr != YADL_FAT_LAST);
   }

   // Remove this song from the playlists
   sb->state= 0;
   cleanup_playlist(0);
      
   // Nullify songbase entry
   memset(sb, 0, sizeof(*sb));

   // All done
}
   

//
//	Usage
//

#define NL "\n"

void 
usage() {
   pr_out= stdout;
   pr("ylink: Yampp MP3 player link tool, version " VERSION 
      NL
      NL "Copyright (c) 2003 Jim Peters <http://uazu.net/>, all rights reserved.  "
      NL "Released under the GNU GPL version 2; see file COPYING for details."
      NL "See http://uazu.net/yampp/ for latest ylink version, and http://yampp.com/"
      NL "for details of Yampp hardware, firmware and other Yampp-related software."
      NL
      NL "Usage: ylink [options] <command> <args> [ : <command> ... ]"
      NL
      NL "Options:"
      NL "  -d <name>             Use device <name> instead of /dev/ttyUSB0"
      NL "  -v                    Verbose"
      NL
      NL "Simple commands:"
      NL "  :                     Command separator, to execute several commands in one go"
      NL "  ping                  Check that the yampp is on and working; also useful to"
      NL "                          return it to normal operation after a command failure"
      NL "  dir                   List all the songs on the yampp"
      NL "  put <file> ....       Write one or more MP3 files onto the player"
      NL "  get <num> ...         Read one or more songs from the player into the current"
      NL "                          directory by their song numbers"
      NL "  del <num>             Delete one or more songs from the player by song number"
      NL "  plget                 Read the playlists into 'playlist.txt' for editing"
      NL "  plput                 Write the data from 'playlist.txt' back to the player"
      NL "  format <name>         Format the player disk in YADL format with the given"
      NL "                          name; requires that the firmware has already filled" 
      NL "                          in the root sector."
      NL "Advanced commands:"
      NL "  info                  Dump out the player's root sector information"
      NL "  dump                  Dump out the root sector, songbase, FAT and playlist"
      NL "                          information in full, for debugging"
      NL "  zeroroot              Zeros the disk root sector.  This forces the yampp"
      NL "                          firmware to reinitialise the root sector.  This can"
      NL "                          be used before a 'format' to completely reinitialise"
      NL "                          the disk from scratch."
      NL "  upload-fw <bin-file>  Upload firmware first time, using yboot boot-loader"
      NL "                          (boot yampp with buttons 3+4 held down)"
      NL "  reload-fw <bin-file>  Upload new firmware assuming yampp firmware is already"
      NL "                          installed and running"
      NL "  yboot-ping            Ping used when in yboot mode"
      NL "  yboot-esc             Exit from yboot mode, i.e. boot firmware"
      );
   exit(1);
}

#undef NL
   
//
//	Command -> tag mapping
//

typedef struct {
   char *cmd;
   int tag;
   int n_arg;
   int lowlev;
} Command;

Command cmdlist[]= {
   { "upload-fw", 'fw0', 1, 1 },
   { "reload-fw", 'fw1', 1, 1 },
   { "yboot-ping", 'ybpi', 0, 1 },
   { "yboot-esc", 'ybes', 0, 1 },
   { "ping", 'ping', 0 },
   { "dir", 'dir', 0 },
   { "get", 'get', -1 },
   { "put", 'put', -1 },
   { "del", 'del', -1 },
   { "plget", 'plge', 0 },
   { "plput", 'plpu', 0 },
   { "format", 'fmt', 1 },
   { "info", 'info', 0 },
   { "dump", 'dump', 0 },
   { "zeroroot", 'zero', 0 },
   { 0, 0 }
};

static Command *
find_cmd(char *str) {
   int a;
   for (a= 0; cmdlist[a].cmd; a++) 
      if (0 == strcmp(str, cmdlist[a].cmd)) 
	 return cmdlist+a;
   return 0;
}


static void NYI() { fprintf(stderr, "Not yet implemented\n"); }

//
//	Main routine
//

char *dev= "/dev/ttyUSB0";

int 
main(int ac, char **av) {
   int a, cmd;
   int loadall= 0;
   int n_cmd= 0;
   char **avnext;
   Command *cc;

#define LOADALL if (!loadall) { loadall= 1; load_all(); }

   ac--, av++;
   verb= 0;

   while (ac > 0 && av[0][0] == '-') {
      char ch, *p= &av[0][1];
      ac--; av++;
      while ((ch= *p++)) switch (ch) {
       case 'v':
	  verb= 1;
	  break;
       case 'd':
	  if (ac < 1) usage();
	  dev= *av++; ac--;
	  break;
       default:
	  usage();
      }
   }

   if (ac == 0) usage();

   // Search for command-separators
   n_cmd= 1;
   for (a= 0; a<ac; a++) {
      if (0 == strcmp(av[a], ":")) {
	 av[a]= 0;
	 n_cmd++;
      }
   }

   // Check for low-level command and handle it; these can't be mixed
   // with other commands
   cc= find_cmd(av[0]);
   if (cc && cc->lowlev) {
      if (n_cmd != 1)
	 error("Low-level commands must be used alone (i.e. no ':')");
      ac--, av++;
      if (ac != cc->n_arg)
	 error("Wrong number of arguments for command \"%s\"", av[-1]);

      // Make connection to the device
      port_open(dev);
      pr_out= stdout;
      
      // Handle certain low-level commands that have special
      // linkup/linkdown requirements
      cmd= cc->tag;
      switch (cmd) {
       case 'fw0':
       case 'fw1':
	  {
	     int len;
	     char *p, *p0, *p1, *p2;
	     char *dat0= load_file(av[0], &len);
	     char *dat= Alloc(len + 128);
	     memcpy(dat, dat0, len); free(dat0);
	     
	     len= (len+127) & ~127;	// Round up to a multiple of 128
	     
	     // Check firmware looks correct
	     p= 0;
	     for (a= 0; a<len-6; a++) 
		if (0 == memcmp("yfwTAG", dat+a, 6)) { p= dat + a; break; }
	     if (!p) error("Firmware file \"%s\" does not contain the 'yfwTAG' tag.\n"
			   "  Do you have the right file?", av[0]);
	     p0= p += 7;
	     p1= p += strlen(p) + 1; 
	     p2= p += strlen(p) + 1; 
	     pr("Firmware file: %s", av[0]);
	     pr("  %s, version %s", p0, p2);
	     pr("  %s", p1);
	     
	     // Go into bootloader mode if we are currently in player mode
	     if (cmd == 'fw1') {
		yampp_link_up();
		yampp_ucb("I", USB_ENTER_BOOT);
	     }
	     
	     // Check that we are in bootloader mode
	     pr("YBOOT bootloader version: %s", yboot_status());
	     
	     pr("Downloading %d 128-byte sectors ...", len/128);
	     yboot_download(dat, len);
	     free(dat);
	     
	     //pr("Pausing ...");
	     //sleep(1);
	     
	     // Return to normal mode, i.e. boot the new firmware
	     pr("Booting new firmware ...");
	     yboot_esc();
	     
	     pr("Pausing ...");
	     sleep(2);
	  }
	  break;
       case 'ybpi':
	  pr("YBOOT version: %s", yboot_status());
	  break;
       case 'ybes':
	  yboot_esc();
	  break;
       default:
	  error("Internal error; low-level command not handled: %s", av[-1]);
      }
      return 0;
   }      

   // Make connection to the device
   port_open(dev);
   pr_out= stdout;
   yampp_link_up();

   // Run through the commands one by one
   avnext= av;
   while (n_cmd-- > 0) {
      av= avnext;
      if (av[0] == 0) { av++; continue; }
      cc= find_cmd(*av++);
      if (!cc) 
	 error("Command-name not known: %s", av[-1]);
      if (cc->lowlev)
	 error("Low-level commands must be used alone: %s", av[-1]);
      
      ac= 0;
      for (avnext= av; *avnext; avnext++) ac++;
      avnext++;

      // ac == number of arguments to this command
      // av[...] == arguments
      // avnext == av for next loop

      if ((cc->n_arg >= 0 && ac != cc->n_arg) ||
	  (cc->n_arg < 0 && ac < -cc->n_arg))
	 error("Wrong number of arguments for command \"%s\"", av[-1]);

      // Handle other commands
      cmd= cc->tag;
      switch (cmd) {
       case 'ping':
	  yampp_ucb("II", USB_TEST_POLL, 'a');
	  if (ACK != yampp_read_rc())
	     error("USB_TEST_POLL failed");
	  printf("Success\n");
	  break;
       case 'dir':
	  LOADALL;
	  list_songs();
	  break;
       case 'get':
	  LOADALL;
	  while (ac-- > 0) 
	     get_song(intarg(*av++));
	  break;
       case 'put':
	  LOADALL;
	  while (ac-- > 0) {
	     put_song(*av++);
	     save_all();	// Save after each one in case of problems
	  }
	  break;
       case 'del':
	  LOADALL;
	  while (ac-- > 0) 
	     del_song(intarg(*av++));
	  save_all();
	  break;
       case 'plge':
	  LOADALL;
	  write_playlist();
	  break;
       case 'plpu':
	  LOADALL;
	  read_playlist();
	  save_all();
	  break;
       case 'fmt':
	  format_disk(av[0]);
	  break;
       case 'info':
	  load_root_raw();
	  pr("Yampp root sector:");
	  dump_root();
	  break;
       case 'dump':
	  load_root_raw();
	  if (!root_valid) {
	     pr("Yampp root sector:");
	     dump_root();
	     break;
	  }
	  load_all();
	  pr("Yampp root sector:");
	  dump_root();
	  pr("");
	  pr("Yampp songbase:");
	  dump_songbase();
	  pr("");
	  pr("Yampp FAT:");
	  dump_fat();
	  pr("");
	  dump_playlist();
	  break;
       case 'zero':
	  memset(&root, 0, sizeof(root));
	  save_root(1);
	  break;
       default:
	  error("Internal error");
      }
   }

   // OLD TESTING STUFF
   // char tmp[16384];
   // { int a; for (a= 0; a<1; a++) { yampp_ucb("III", USB_SECTOR_READ, a, 1); } }
   // yampp_ucb("II", USB_TEST_POLL, 'b');
   // yampp_ucb("I", USB_IDENTIFY);
   // yampp_ucb("I", USB_BUFFERINFO);
   // port_read(tmp, sizeof(tmp));

   // Link down
   yampp_link_down();

   return 0;
}

// END //
